#########################################
## THIS FILE IS PART OF BRAINGENIX     ##
## COMPUTER NEURON SIMULATUION LIBRARY ##
#########################################

import random
import pickle

import numpy as np



## RUNS A NEURON GROUP ##

class NG(): # is a neuron group

    def __init__(self, ndim_w = 5, ndim_h = 5, ndim_d = 5, mcons = 5000, tick_speed = 0.5, rp = -70, tp = -55, ap = 40, hp = -150, dp_t = 1, rp_t = 1, hp_t = 2, ncondist_max = 3): # initalizes the processing matrix


        '''
        NOTE:
        ALL TIMES ARE IN ms
        ALL VOLAGES ARE IN mv
        TIMES MUST BE SYNCED TO ALL THREADS TO PREVENT UNEXPECTED RESULTS
        '''



        ## INITALIZE SIMULATION PARAMETERS LATER DOWN THE PIPELINE ##

        #self.input_queue = input_queue # set the input queue

        #self.output_queue = output_queue # set the output queue


        self.neuron_width = ndim_w # set the neuron width of the NG

        self.neuron_height = ndim_h # set the neuron height of the NG

        self.neuron_depth = ndim_d # set the neuron depth of the NG


        self.neuron_threshold_potential = tp # set the threshold voltage

        self.action_potential = ap # set the action potential voltage

        self.hyper_polarization = hp # set the hyper polarization voltage
       
        self.resting_potential = rp # set the neuron resting potential voltage

        
        self.depolarazation_time = dp_t # set depolarazation time

        self.repolarazation_time = rp_t # set repolarazation time

        self.hyper_polarization_time = hp_t # set hyperpolarazation time

        
        self.neuron_ticks_speed = tick_speed # set the tick speed for the simulation

        self.neuron_connection_distance_max = ncondist_max # set the max neuron connection distance

        self.max_connections = mcons # set the maximum number of connections allowed 

        self.weight_increase = 0.001 # set the neuron weight increase


        ## INITALZIE THE NEURON VOLTAGES ##

        resting_potential = rp # set the resting potential

        self.neuron_voltage = [] # make the action potentials array

        for z in range(ndim_w * ndim_h * ndim_d): # iterate through the number of desired neurons

            self.neuron_voltage.append(resting_potential) # set all neurons to the resting potential


        ## INITALIZE NEURON TYPE ##

        self.neuron_types = [] # make an empty array

        '''NOTE: 1 is exitory, and 0 is inhibitory
        '''

        for z in range(ndim_w * ndim_h * ndim_d): # iterate through the number of desired neurons

            self.neuron_types.append(random.randint(0,1)) # set the neuron type

            

        ## INITALIZE THE CONNECTION ARRAY ##

        self.connections = [] # set inital connections array

        for z in range(ndim_w * ndim_h * ndim_d): # iterate through the number of desired neurons

            self.connections.append([[],[]]) # initalize every NG with no connecitons



        ## INITALIZE NEURON STAGES (FOR DEPOLARAZATION AS WELL AS HYPERPOLARAZATION, ETC) ##

        self.neuron_stages = [] # set inital stage array

        for z in range(ndim_w * ndim_h * ndim_d): # iterate through all the desired stage arrays

            self.neuron_stages.append(0) # append the default stage index



        ## CALCULATE THE NEURON STAGE VOLTAGE VALUES ##

        self.neuron_stage_values = [] # set inital array

        for z in range(int(dp_t / self.neuron_ticks_speed)): # make the first part of the depolarazation curve

            slices = (self.depolarazation_time / self.neuron_ticks_speed) - 1 # calculate the number of slices needed

            v_diff = (abs(self.action_potential) + abs(self.resting_potential)) # set the voltage difference


            voltage = (v_diff / slices) * z + self.resting_potential # calculate the voltage

            self.neuron_stage_values.append(voltage) # append the voltage to the array


        for z in range(int(rp_t / self.neuron_ticks_speed)): # calculate the repolarazation curve

            slices = (self.repolarazation_time / self.neuron_ticks_speed) - 1 # calculate the number of slices needed

            v_diff = -(abs(self.action_potential) + abs(self.resting_potential)) # set the voltage difference


            voltage = (v_diff / slices) * z + self.action_potential # calculate the voltage

            self.neuron_stage_values.append(voltage) # append the voltage to the array


        for z in range(int((hp_t / 2) / self.neuron_ticks_speed)): # calculate the repolarazation curve

            slices = ((self.hyper_polarization_time / 2) / self.neuron_ticks_speed) - 1 # calculate the number of slices needed

            v_diff = (abs(self.resting_potential) - abs(self.hyper_polarization)) # set the voltage difference


            voltage = (v_diff / slices) * z + self.resting_potential # calculate the voltage

            self.neuron_stage_values.append(voltage) # append the voltage to the array


        for z in range(int((hp_t / 2) / self.neuron_ticks_speed)): # calculate the repolarazation curve

            slices = ((self.hyper_polarization_time / 2) / self.neuron_ticks_speed) - 1 # calculate the number of slices needed

            v_diff = (abs(self.resting_potential) - abs(self.hyper_polarization)) # set the voltage difference


            voltage = -((v_diff / slices) * z - self.hyper_polarization) # calculate the voltage

            self.neuron_stage_values.append(voltage) # append the voltage to the array



        ## NEXT, MAKE THE INTERCONNECT LINKS FOR THE TOP, BOTTOM, RIGHT, LEFT, FRONT, AND BACK ##

        self.interconnect_top_base = [] # make the inital top interconnect link

        self.interconnect_bottom_base = [] # make the inital bottom interconnect link

        self.interconnect_left_base = [] # make the inital left interconnect link

        self.interconnect_right_base = [] # make the inital right interconnect link

        self.interconnect_front_base = [] # make the inital front interconnect link

        self.interconnect_back_base = [] # make the inital back interconnect link


        for z in range(self.neuron_depth * self.neuron_width): # populate the top interconnect layer with an empty list

            self.interconnect_top_base.append(0) # populate the top interconnect layer

            self.interconnect_bottom_base.append(0) # populate the lower interconnect layer


        for z in range(self.neuron_depth * self.neuron_height): # populate the top interconnect layer with an empty list

            self.interconnect_left_base.append(0) # populate the left interconnect layer

            self.interconnect_right_base.append(0) # populate the right interconnect layer


        for z in range(self.neuron_width * self.neuron_height): # populate the top interconnect layer with an empty list

            self.interconnect_front_base.append(0) # populate the front interconnect layer

            self.interconnect_back_base.append(0) # populate the back interconnect layer


        ## ONCE DONE INITALIZING, LOOP AND CALL RUN ##

        #while True: # forever

            #self.process_tick() # call the process tick


    def process_tick(self, input_link_data): # process the neuron group for this tick

        '''
        NOTE: THE DATA FORMAT IS AS FOLLOWS

        [[Front], [Back], [Top], [Bottom], [Left], [Right]]

        '''

        ## INTITALIZE THE NEURON-GROUP TOTAL VOLTAGE ##

        ng_total_voltage = 0 # initalize the total voltage to be 0


        ## GET DATA FROM THE QUEUE ##

        #input_link_data = self.input_queue.get() # get the data from the queue


        ## CHECK IF IT'S A COMMAND ##

        if type(input_link_data) == str: # if it's a string

            data = input_link_data.split('&') # split the command from the args by a "&"

            cmd = data[0] # get the command aspect


            if cmd == 'SAVE': # if we're to save

                self.save_ng(str(data[1])) # save the neuron GROUP


            if cmd == 'LOAD': # if we're to load

                self.load_ng(str(data[1])) # load the neuron GROUP


            elif cmd == 'EXIT': # if we're to exit

                return 0 # exit the process by returning a value


        else: # otherwise



            ## RESET THE NEURON VOLTAGES ##

            self.neuron_voltage = [] # make the action potentials array

            for z in range(self.neuron_width * self.neuron_height * self.neuron_depth): # iterate through the number of desired neurons

                    self.neuron_voltage.append(self.resting_potential) # set all neurons to the resting potential


            ## NEXT THING TO DO, IS TO FEED IN THE INPUT DATA INTO THE NEURON GROUP ##

            if input_link_data != [[],[],[],[],[],[]]: # if has some input content



                ## MAP TO NEURONS (FRONT LAYER) ##

                if input_link_data[0] != []: # if the front layer isn't empty

                    for z in range(self.neuron_width * self.neuron_height): # iterate through the front face of the neuron to feed in the data

                        index = (z * self.neuron_depth) + self.neuron_depth - 1 # calculate the index to update the input voltage to

                        self.neuron_voltage[index] = input_link_data[0][z] # update the input layer's data




                ## MAP TO NEURONS (BACK LAYER) ##

                if input_link_data[1] != []: # if the front layer isn't empty

                    for z in range(self.neuron_width * self.neuron_height): # iterate through the front face of the neuron to feed in the data

                            index = (z * self.neuron_depth) # calculate the index to update the input voltage to

                            self.neuron_voltage[index] = input_link_data[1][z] # update the input layer's data


                ## MAP TO NEURONS (TOP LAYER) ##

                if input_link_data[2] != []: # if the front layer isn't empty

                    for z in range(self.neuron_width * self.neuron_depth): # iterate through the front face of the neuron to feed in the data

                        index = (self.neuron_width * self.neuron_depth * (self.neuron_height - 1)) + z # calculate the index to update the input voltage to

                        self.neuron_voltage[index] = input_link_data[2][z] # update the input layer's data



                ## MAP TO NEURONS (BOTTOM LAYER) ##

                if input_link_data[3] != []: # if the front layer isn't empty

                    for z in range(self.neuron_width * self.neuron_depth): # iterate through the front face of the neuron to feed in the data

                        index = z # calculate the index to update the input voltage to

                        self.neuron_voltage[index] = input_link_data[3][z] # update the input layer's data



                ## MAP TO NEURONS (LEFT LAYER) ##

                if input_link_data[4] != []: # if the front layer isn't empty

                    for z in range(self.neuron_width * self.neuron_depth): # iterate through the front face of the neuron to feed in the data

                        index = (int(self.neuron_height / 5) * (self.neuron_width - 1) * self.neuron_depth) + z # calculate the index to update the input voltage to

                        self.neuron_voltage[index] = input_link_data[4][z] # update the input layer's data


                ## MAP TO NEURONS (RIGHT LAYER) ##

                if input_link_data[5] != []: # if the front layer isn't empty

                    for z in range(self.neuron_width * self.neuron_depth): # iterate through the front face of the neuron to feed in the data

                        index = ((self.neuron_width - 1) * self.neuron_depth) + (int(self.neuron_height / 5) * (self.neuron_width - 1) * self.neuron_depth) + z # calculate the index to update the input voltage to

                        self.neuron_voltage[index] = input_link_data[5][z] # update the input layer's data


        
            ## DEAL WITH THE ACTUAL NEURON PROCESSING ##

            for neuron_index in range(len(self.neuron_voltage)): # iterate through the neurons


                ## FIRST, SEE IF THE NEURON IS TO DEPOLARIZE ##

                if (self.neuron_voltage[neuron_index] >= self.neuron_threshold_potential) and (self.neuron_stages[neuron_index] == 0): # if the neuron is to depolarize


                    ## INCRIMENT THE NEURON STAGE COUNTER UP BY 1 ##

                    self.neuron_stages[neuron_index] += 1 # incriment up the stage counter

                
                    ## ATTEMPT TO MAKE A NEW CONNECTION ##

                    if random.randint(0, 4) == 1: # if it's to make a new connection


                        ############################## START COMMENT OUT ##################################

                        #near_indexes = [] # make a list of near indexes to the neuron

                        #neuron_layer = int(neuron_index / (self.neuron_depth * self.neuron_width)) # calculate the neuron's layer


                        ### MAKE A LIST OF THE LAYERS TO BUILD THE CONNECTION ARRAY ON ##

                        #neuron_layers = [] # make a new list for the neurons to populate on

                        #for x in range(self.neuron_connection_distance_max): # iterate through the connection array

                            #nlayer = neuron_layer - self.neuron_connection_distance_max

                            #if (nlayer > -1) and (nlayer < self.neuron_height): # if it's in range

                                #neuron_layers.append(nlayer) # append the nlayer index


                        ## PROCESS ALL LAYERS AND MAKE A LIST OF POSSIBLE INDEXES ##

                        #indexes = [] # make list to store indexes

                        #for layer in neuron_layers: # iterate through the possible neuron layers

                            #base_index = layer * (self.neuron_depth * self.neuron_width) # get the base layer

                            ############################ END COMMENT OUT ##################################
                
                        if len(self.connections[neuron_index][0]) < self.max_connections: # if it hasn't exceeded the max number of connections


                            ## MAKE A CONNECTION TO ANOTHER NEURON OR INTERCONNECT POINT ##

                            weight = random.randint(0, 100000) / 100000 # generate a random weight
                        
                            connection_index = random.randint(0, len(self.neuron_voltage) - 1) # generate a random connection index


                            ## CHECK TO MAKE SURE THAT THE INDEX WILL WORK ##

                            valid = False # set inital value

                            if (connection_index != neuron_index) and (connection_index not in self.connections[connection_index][0]): # if its valid

                                valid = True # use it as a good connection



                            if valid: # if it's valid

                            
                                ## APPEND IT ##

                                self.connections[neuron_index][0].append(connection_index) # append the connection indexes

                                self.connections[neuron_index][1].append(weight) # append the connection weights


                ## NEXT, HANDLE THE CONNECTIONS AND THEIR WEIGHTS ##

                current_voltage = self.neuron_stage_values[self.neuron_stages[neuron_index]] # get the current volage on that neuron

            
                ## CHECK TO SEE IF IT HAS ANY CONNECTIONS MADE FIRST ##

                if self.connections[neuron_index] != [[],[]]: # check to see that it actually has conencitons

                    ## UNPACK THE NEURON DATA VALUES ##

                    synapses, weights = self.connections[neuron_index] # unpack the synapse index data

                    for synapse_index in range(len(self.connections[neuron_index][0])): # iterate through the connections


                        ## UPDATE THE NEURON'S VOLTAGE ##

                        synapse = synapses[synapse_index] # get the synapses

                        weight = weights[synapse_index] # get the weights


                        voltage_addition = weight * current_voltage # calculate the voltage difference between them

                        old_neuron_voltage = self.neuron_voltage[synapse] # get the old voltage


                        ## CHECK NEURON TYPE ##

                        if self.neuron_types[neuron_index] == 1: # if it's exitory 

                            new_neuron_voltage = old_neuron_voltage + voltage_addition # add the two voltages

                        else: # if it's inhibitory

                            new_neuron_voltage = old_neuron_voltage - voltage_addition # subtract the two voltages

                        self.neuron_voltage[synapse] = new_neuron_voltage # update the neuron voltage on that index


                        ## INCREASE THE CONNECTION'S STRENGTH ##

                        conn_weight = weight + self.weight_increase # incriment up the connection weight so that it will fire more in the future

                        self.connections[neuron_index][1][synapse_index] = conn_weight # update the connection weight
                            

                ## PROCESS UPDATING THE NEURON STAGES ##

                if self.neuron_stages[neuron_index] != 0: # if it's depolarizing

                    self.neuron_stages[neuron_index] = (self.neuron_stages[neuron_index] + 1) % len(self.neuron_stage_values) # get the length of the neuron stages


                ## UPDATE THE TOTAL NEURON GROUP VOLTAGE VARIABLE ##

                ng_total_voltage += self.neuron_stage_values[self.neuron_stages[neuron_index]] # update the total variable





                
            ## COMPILE THE NEURON OUTPUT LAYERS AND RETURN THEM TO THE MAIN PROCESS ##

            interconnect_layer_output = [] # make an array for neuron output interconnects


            ## MAP TO NEURONS (FRONT LAYER) ##

            front_out = [] # make a new output array

            for z in range(self.neuron_width * self.neuron_height): # iterate through the front face of the neuron to feed in the data

                index = (z * self.neuron_depth) + self.neuron_depth - 1 # calculate the index to feed the output voltage to

                front_out.append(self.neuron_voltage[index]) # get the input layer's data and append it to the output layer



            ## MAP TO NEURONS (BACK LAYER) ##

            back_out = [] # make new output array

            for z in range(self.neuron_width * self.neuron_height): # iterate through the front face of the neuron to feed in the data

                index = (z * self.neuron_depth) # calculate the index to feed the output voltage to

                back_out.append(self.neuron_voltage[index]) # get the input layer's data and append it to the output layer



            ## MAP TO NEURONS (TOP LAYER) ##

            top_out = [] # make new output array

            for z in range(self.neuron_width * self.neuron_depth): # iterate through the front face of the neuron to feed in the data

                index = (self.neuron_width * self.neuron_depth * (self.neuron_height - 1)) + z # calculate the index to feed the output voltage to

                top_out.append(self.neuron_voltage[index]) # get the input layer's data and append it to the output layer



            ## MAP TO NEURONS (BOTTOM LAYER) ##

            bottom_out = [] # make new output array

            for z in range(self.neuron_width * self.neuron_depth): # iterate through the front face of the neuron to feed in the data

                index = z # calculate the index to feed the output voltage to

                bottom_out.append(self.neuron_voltage[index]) # get the input layer's data and append it to the output layer



            ## MAP TO NEURONS (LEFT LAYER) ##

            left_out = [] # make new output array

            for z in range(self.neuron_width * self.neuron_depth): # iterate through the front face of the neuron to feed in the data

                index = (int(z / self.neuron_height) * (self.neuron_width - 1) * self.neuron_depth) + z # calculate the index to feed the output voltage to

                left_out.append(self.neuron_voltage[index]) # get the input layer's data and append it to the output layer



            ## MAP TO NEURONS (RIGHT LAYER) ##

            right_out = [] # make new output array

            for z in range(self.neuron_width * self.neuron_depth): # iterate through the front face of the neuron to feed in the data

                index = ((self.neuron_width - 1) * self.neuron_depth) + (int(z / self.neuron_height) * (self.neuron_width - 1) * self.neuron_depth) + z # calculate the index to feed the output voltage to

                right_out.append(self.neuron_voltage[index]) # get the input layer's data and append it to the output layer



            ## COMPILE THE OUTPUT INFORMATION ##

            output_data = [front_out, back_out, top_out, bottom_out, left_out, right_out] # make an output array


            return [output_data, ng_total_voltage] # return the data


            #self.output_queue.put([output_data, ng_total_voltage]) # return all of the data in the queue




    def save_ng(self, filename = 'save.dvn'): # saves the AI's state


        
##        ## FIRST, MAKE THE SAVE DATA VARIABLE ##
##
##        save_data = ''  # make the save data variable
##
##
##        ## NEXT, COMPILE THE SETTINGS INTO IT ##
##
##        save_data += '{}*{}*{}*{}*{}*{}*{}*{}*{}*{}*'.format(str(self.neuron_width), str(self.neuron_height), str(self.neuron_depth), str(self.neuron_threshold_potential), str(self.action_potential), str(self.hyper_polarization), str(self.resting_potential), str(self.depolarazation_time), str(self.repolarazation_time), str(self.hyper_polarization_time)) # write the simulation paramaters to the save data
##
##        save_data += '{}*{}*{}*{}!1234567890)'.format(str(self.neuron_ticks_speed), str(self.neuron_connection_distance_max), str(self.max_connections), str(self.weight_increase)) # write the simulation parameters to the save data
##
##
##        ## FINALLY SAVE THE IMPORTANT DATA ##
##        
##        #for nv in self.neuron_voltage: # iterate through the neuron voltage array to save all the voltages
##
##            #save_data += '{}*'.format(str(nv)) # write the neuron voltage data
##
##
##
##        save_data += str(self.neuron_voltage).strip("'") # add the data
##
##        save_data = save_data[:-1] # remove the last character of the save_data string
##
##        save_data += '!1234567890)' # add a catagorial split string
##
##
##        for idx in range(len(self.connections)): # iterate through the connections array
##
##            #for connection in self.connections[idx][0]: # iterate through the connections
##
##                #save_data += '{}*'.format(str(connection)) # add the index to the save_data variable  
##
##            save_data += str(self.connections[idx]).strip("'") # add the save data
##                
##            save_data += '#' # add line split
##
##        save_data = save_data[:-1] # remove the last character of the save_data string
##
##        save_data += '!1234567890)' # add a catagorial split string
##
##        #print(len(self.connections))
##        #print('3')
##        #for idx in range(len(self.connections)): # iterate through the connections array
##
##        #    for connection in self.connections[idx][1]: # iterate through the connections
##
##        #        save_data += '{}*'.format(str(connection)) # add the index to the save_data variable 
##                
##        #    save_data += '#' # add line split
##
##        #save_data = save_data[:-1] # remove the last character of the save_data string
##
##        #save_data += '!1234567890)' # add a catagorial split string
##
##
##        #for stage in self.neuron_stages: # iterate through the neuron stage VALUES
##
##            #save_data += '{}*'.format(str(stage)) # append the stage values to the save_data variable 
##
##        save_data += str(self.neuron_stages).strip("'") # add the neuron stages to it
##
##        save_data = save_data[:-1] # remove the last character of the save_data string
##
##        save_data += '!1234567890)' # add a catagorial split string
##
##
##        ## FINALLY WRITE THE DATA TO THE FILE ##
##
##        with open(filename, 'w') as f: # open the file to write to it
##
##            f.write(save_data) # save the file


        ## NEW WAY OF SAVING ARRAY, USE PICKLE ##

        write_array = [] # make an empty list to be dumpted


        write_array.append(self.neuron_width) # append it

        write_array.append(self.neuron_height) # append it 

        write_array.append(self.neuron_depth) # append it

        write_array.append(self.neuron_threshold_potential) # append it 

        write_array.append(self.action_potential) # append it

        write_array.append(self.hyper_polarization) # append it

        write_array.append(self.resting_potential) # append it

        write_array.append(self.depolarazation_time) # append it 

        write_array.append(self.repolarazation_time) # append it 

        write_array.append(self.hyper_polarization_time) # append it 


        write_array.append(self.neuron_ticks_speed) # append it

        write_array.append(self.neuron_connection_distance_max) # append it

        write_array.append(self.max_connections) # append it 

        write_array.append(self.weight_increase) # append it



        write_array.append(self.neuron_voltage) # append it

        write_array.append(self.connections) # append it

        write_array.append(self.neuron_stages) # append it


        ## FINALLY WRITE THE DATA TO THE FILE ##

        with open(filename, 'wb') as f: # open the file to write to it

            pickle.dump(write_array, f) # save the file
        


    def load_ng(self, filename = 'save.dvn'): # saves the AI's state


        ## OLD METHOD OF DOING IT, COMMENTED OUT DUE TO DISUSE ##
        
        
##        ## FIRST, OPEN THE FILE AND READ THE CONTENTS FOR PARSING ##
##
##        with open(filename, 'r') as f: # open the file to READ
##
##            data = f.read() # read the file
##
##
##        ## NEXT, PARSE IT ##
##
##        data = data.split('!1234567890)') # split it based upon the catagorial split string
##
##
##        ## LOAD THE SIMULATION PARAMATERS ##
##
##        parameters = data[0].split('*') # get the paramater data
##
##
##        self.neuron_width = parameters[0] # extract the data
##
##        self.neuron_height = parameters[1] # extract the data
##
##        self.neuron_depth = parameters[2] # extract the data
##
##        
##        self.neuron_threshold_potential = parameters[3] # extract the data
##
##        self.action_potential = parameters[4] # extract the data
##
##        self.hyper_polarization = parameters[5] # extract the data
##
##        self.resting_potential = parameters[6] # extract the data
##
##
##        self.depolarazation_time = parameters[7] # extract the data
##
##        self.repolarazation_time = parameters[8] # extract the data
##
##        self.hyper_polarization_time = parameters[9] # extract the data
##
##
##        self.neuron_ticks_speed = parameters[10] # extract the data
##
##        self.neuron_connection_distance_max = parameters[11] # extract the data
##
##        self.max_connections = parameters[12] # extract the data
##
##        self.weight_increase = parameters[13] # extract the data
##
##
##        ## NEXT, UNLOAD THE NEURON_VOLTAGE DATA ##
##
##        nvd = data[1].strip('[').strip(']').split(', ') # load the neuron_voltage data
##        
##        self.neuron_voltage = list(map(float, nvd)) # convert to float, then update list data
##
##        
##        ## THEN UNLOAD THE CONNECTION DATA ##
##
##        conn_indexes = data[2].replace('[', '').replace(']', '').split(', ') # load the connection INDEXES
##
##        conn_weights = data[3].replace('[', '').replace(']', '').split(', ') # load the connection WEIGHTS
##
##
##        print(conn_indexes)
##
##
##        self.connections = [list(map(int, conn_indexes)), list(map(float, conn_weights))] # convert list types, update connections array
##
##
##        ## NEXT, UNLOAD THE NEURON STAGE DATA ##
##
##        stages = data[4].split('*') # unload the data and split it
##
##        self.neuron_stages = list(map(int, stages)) # format and update the neuron_stages list


        ## PICKLE METHOD OF SAVE READING, MUCH FASTER ##



        ## OPEN THE PICKLE DUMP ##

        data = pickle.load(open(filename, 'rb')) # read the data


        ## UNPACK THE DATA ##

        self.neuron_width = data[0] # get the data

        self.neuron_height = data[1] # get the data

        self.neuron_depth = data[2] # get the data

        self.neuron_threshold_potential = data[3] # get the data

        self.action_potential = data[4] # get the data

        self.hyper_polarization = data[5] # get the data

        self.resting_potential = data[6] # get the data

        self.depolarazation_time = data[7] # get the data

        self.repolarazation_time = data[8] # get the data

        self.hyper_polarization_time = data[9] # get the data

        self.neuron_ticks_speed = data[10] # get the data

        self.neuron_connection_distance_max = data[11] # get the data

        self.max_connections = data[12] # get the data

        self.weight_increase = data[13] # get the data


        self.neuron_voltage = data[14] # get the data

        self.connections = data[0] # get the data

        self.neuron_stages = data[0] # get the data




        ## NEXT, RECALCULATE THE 'neuron_stage_values' ARRAY ##

        self.neuron_stage_values = [] # set inital array

        for z in range(int(self.depolarazation_time / self.neuron_ticks_speed)): # make the first part of the depolarazation curve

            slices = (self.depolarazation_time / self.neuron_ticks_speed) - 1 # calculate the number of slices needed

            v_diff = (abs(self.action_potential) + abs(self.resting_potential)) # set the voltage difference


            voltage = (v_diff / slices) * z + self.resting_potential # calculate the voltage

            self.neuron_stage_values.append(voltage) # append the voltage to the array


        for z in range(int(self.repolarazation_time / self.neuron_ticks_speed)): # calculate the repolarazation curve

            slices = (self.repolarazation_time / self.neuron_ticks_speed) - 1 # calculate the number of slices needed

            v_diff = -(abs(self.action_potential) + abs(self.resting_potential)) # set the voltage difference


            voltage = (v_diff / slices) * z + self.action_potential # calculate the voltage

            self.neuron_stage_values.append(voltage) # append the voltage to the array


        for z in range(int((self.hyper_polarization_time / 2) / self.neuron_ticks_speed)): # calculate the repolarazation curve

            slices = ((self.hyper_polarization_time / 2) / self.neuron_ticks_speed) - 1 # calculate the number of slices needed

            v_diff = (abs(self.resting_potential) - abs(self.hyper_polarization)) # set the voltage difference


            voltage = (v_diff / slices) * z + self.resting_potential # calculate the voltage

            self.neuron_stage_values.append(voltage) # append the voltage to the array


        for z in range(int((self.hyper_polarization_time / 2) / self.neuron_ticks_speed)): # calculate the repolarazation curve

            slices = ((self.hyper_polarization_time / 2) / self.neuron_ticks_speed) - 1 # calculate the number of slices needed

            v_diff = (abs(self.resting_potential) - abs(self.hyper_polarization)) # set the voltage difference


            voltage = -((v_diff / slices) * z - self.hyper_polarization) # calculate the voltage

            self.neuron_stage_values.append(voltage) # append the voltage to the array



        ## FINALLY, MAKE THE INTERCONNECT LINKS FOR THE TOP, BOTTOM, RIGHT, LEFT, FRONT, AND BACK ##

        self.interconnect_top_base = [] # make the inital top interconnect link

        self.interconnect_bottom_base = [] # make the inital bottom interconnect link

        self.interconnect_left_base = [] # make the inital left interconnect link

        self.interconnect_right_base = [] # make the inital right interconnect link

        self.interconnect_front_base = [] # make the inital front interconnect link

        self.interconnect_back_base = [] # make the inital back interconnect link


        for z in range(self.neuron_depth * self.neuron_width): # populate the top interconnect layer with an empty list

            self.interconnect_top_base.append(0) # populate the top interconnect layer

            self.interconnect_bottom_base.append(0) # populate the lower interconnect layer
            

        for z in range(self.neuron_depth * self.neuron_height): # populate the top interconnect layer with an empty list

            self.interconnect_left_base.append(0) # populate the left interconnect layer

            self.interconnect_right_base.append(0) # populate the right interconnect layer


        for z in range(self.neuron_width * self.neuron_height): # populate the top interconnect layer with an empty list

            self.interconnect_front_base.append(0) # populate the front interconnect layer

            self.interconnect_back_base.append(0) # populate the back interconnect layer

