import time
import multiprocessing
import cv2
import matplotlib
import os
import random
import datetime

import numpy as np
import matplotlib.pyplot as plt

from BrainGenix.multiprocessing.Neuron_Groups import NG
from BrainGenix.core.Logger import Error_Logger
from BrainGenix.core.Config_Reader import Read_Config
from BrainGenix.core.Video_Captures import videocapture
from BrainGenix.core.FileIO import trymake

class asi(): # is the class that runs the entire ai

        def __init__(self, log_location = 'log.txt', config_path = 'system.prefs', save_dir = 'saves', graph_path = 'graphs'): # initalizes everything



                ## DEFINE SAVE DIR VAR ##


                self.save_dir = save_dir # set the save dir var

                self.graph_path = graph_path # set the graph path var
                

                ## FIRST, INITALIZE THE ERROR LOGGER ##

                print('Initalizing Error Logger. (NOTE: THIS MSG WILL NOT BE IN THE LOG FILE)') # print startup message

                self.e_log = Error_Logger(msg_per_write = 10, fname = log_location) # initalize the error logger

                self.e_log.print_log('Logging System Initalized') # info stuff



                ## DETECT AVAILABLE CPU COUNT ##

                self.cpu_count = multiprocessing.cpu_count() # get number of cpus

                self.e_log.print_log('The System Will Use {} Cpus Unless Otherwise Specified'.format(str(self.cpu_count))) # info stuff



                ## LOAD THE CONFIG FILE ##

                self.prefs = Read_Config(config_path) # read the config file

                self.e_log.print_log('Loaded Config File') # info stuff


                ## CHECK IF A PRIOR MODEL EXISTS ## 

                target_str = 'DIMS-{}-{}-{}/'.format(self.prefs[3], self.prefs[4], self.prefs[5]) # get the target string

                prev_models = os.listdir(save_dir) # list all the models in the saves directory


                applicable_models = [] # make an empty list

                for teststr in applicable_models: # iterate through the teststring to check if we can use this string


                        if target_str in teststr: # if it's its usable

                                applicable_models.append(teststr) # append it


                max_version_number = 0 # set the default version number

                best_model = 0 # set the best index

                for model in applicable_models: # iterate through the test model name

                        version = model.strip('AI_NEURAL_NET-') # remove the junk data

                        version = int(version.split('-')[0]) # get the verion number

                        if verison > max_version_number: # if its newer

                                max_version_number = version # set the new v num

                                best_model = model # set it




                ## INITALIZE THE INPUT QUEUES (RELATIVE TO THE WORKER PROCS) ##

                self.e_log.print_log('Initalizing Input Queues') # info stuff

                self.input_queues = [] # make a list of input queues to send data to

                for z in range(int(self.prefs[3]) * int(self.prefs[4]) * int(self.prefs[5])): # get the number of input queues required

                        self.input_queues.append(multiprocessing.Queue()) # initalize the queue

                        self.e_log.print_log('Created Input Queue {}'.format(str(z))) # info stuff

                self.e_log.print_log('Successfully Initalized Input Queues') # info stuff



                ## INITALIZE THE OUTPUT QUEUES (RELATIVE TO THE WORKER PROCS) ##

                self.e_log.print_log('Initalizing Output Queues') # info stuff

                self.output_queues = [] # make a list of input queues to send data to

                for z in range(int(self.prefs[3]) * int(self.prefs[4]) * int(self.prefs[5])): # get the number of output queues required

                        self.output_queues.append(multiprocessing.Queue()) # initalize the queue

                        self.e_log.print_log('Created Output Queue {}'.format(str(z))) # info stuff

                self.e_log.print_log('Successfully Initalized Output Queues') # info stuff



                ## INITALIZE THE WORKER PROCESSES ##

                self.e_log.print_log('Initalizing Worker Processes') # info stuff

                self.workers = [] # make a list to hold the worker processes

                for z in range(int(self.prefs[3]) * int(self.prefs[4]) * int(self.prefs[5])): # get the number of processes required

                        self.workers.append(multiprocessing.Process(target = NG, args = (self.input_queues[z], self.output_queues[z], int(self.prefs[0]), int(self.prefs[1]), int(self.prefs[2]), int(self.prefs[6]), float(self.prefs[7]), ))) # spawn the worker process

                        self.e_log.print_log('Spawned Worker Process {}'.format(str(z))) # info stuff

                self.e_log.print_log('Successfully Initalized Worker Processes') # info stuff



                ## START THE WORKER PROCESSES ##

                self.e_log.print_log('Starting Worker Processes') # info stuff

                for z in range(int(self.prefs[3]) * int(self.prefs[4]) * int(self.prefs[5])): # get the number of processes required

                        self.workers[z].start() # start the worker process

                        self.e_log.print_log('Started Worker Process {}'.format(str(z))) # info stuff

                self.e_log.print_log('Successfully Started Worker Processes') # info stuff



                ## IF THE SOURCE IS A VIDEO, OPEN THE VIDEO TO START READING FROM IT ##

                if (self.prefs[8] == 'Video') or (self.prefs[8] == 'video'): # if it's a video

                        self.cap = cv2.VideoCapture(self.prefs[9]) # initalize a VideoCapture to the file



                ## IF THE SOURCE IS FROM THE WEBCAM, OPEN IT TO START READING FROM IT ##

                if (self.prefs[8] == 'Webcam') or (self.prefs[8] == 'webcam'): # if it's a video

                        self.cap = videocapture(self.prefs[10], res=(int(self.prefs[3]) * int(self.prefs[0]), int(self.prefs[4]) * int(self.prefs[1]))) # initalize a VideoCapture to the Webcam, on the specified device id

        

                ## CREATE THE INTERCONNECT LINK MAP ##

                '''Note: This just tells it what NG cubes are where, so that it can attach the ai to it'''

                self.index_map = [] # create an index map



                rows = [] # make a row of indexes

                for z in range(int(len(self.workers) / int(self.prefs[5]))): # iterate through the workers list, and come up with the number of required rows

                        tmp = [] # make a tmp list

                        for x in range(int(self.prefs[5])): # iterate through the number of desired indices per row

                                tmp.append((z * int(self.prefs[5])) + x) # append the index

                        rows.append(tmp) # append the rows to the main rows list



                for z in range(int(self.prefs[4])): # iterate through the layers

                        tmp = [] # make a tmp list

                        for x in range(int(self.prefs[3])): # iterate through the width

                                tmp.append(rows[(z * int(self.prefs[3])) + x]) # append the rows

                        self.index_map.append(tmp) # append the main index to the full map

                
                ## INITALIZE VARIABLES ##

                self.tick_counter = 0 # set the current tick to 0

                self.last_tick_interconnect_layers = [] # set the interconnect link layer to an empty list

                self.voltages = [] # make a list for the voltages


                ## POPULATE THE INTERCONNECT LINK LAYERS ##

                for z in range(int(self.prefs[3]) * int(self.prefs[4]) * int(self.prefs[5])): # multiply through the desired number of interconnects

                        self.last_tick_interconnect_layers.append([[],[],[],[],[],[]]) # append an empty list to it
                                


                        

        def run(self): # processes everything

                while True: # loop forever



                        ## SET THE MODE ##

                        mode = 'RUN' # set the defult mode

                        if self.tick_counter % int(self.prefs[13]) == 0: # if it's time to make the graph

                            mode = 'save_graph' # save the graph
                      
                        if self.tick_counter % int(self.prefs[12]) == 0: # if it's time to save

                            mode = 'SAVE' # save the simulation





                        ## GET TICK EXECUTION START TIME ##

                        start_time = time.time() # get the start time



                        ## CHECK IF WE NEED TO FEED THE NEXT FRAME ##

                        if self.tick_counter % 14 == 0: # if it's time to send in the next frame

                                if (self.prefs[8] == 'Video') or (self.prefs[8] == 'Webcam'): # if we need to read a frame

                                        ## FIRST, GET THE NEXT FRAME IN THE DATA TO READ ##

                                        frame = self.cap.read() # reads the next video frame

                                        if frame[0] == False: # if it fails to READ
                                
                                                self.e_log.print_log('Unable To Read Frame! Either File Does Not Exist, Or The System Could Not Fild A Matching Codec.', 4) # info stuff

                                        if (self.prefs[8] == 'Video') or (self.prefs[8] == 'video'): # if it's a video

                                                frame = frame[1] # remove the ret from the data




                                        ## NOW, RESIZE IT TO FIT IN THE INPUT SHAPE ##

                                        target_shape_w = int(int(self.prefs[3]) * int(self.prefs[0]) / 3) # get the desired width

                                        target_shape_h = int(self.prefs[4]) * int(self.prefs[1]) # get the desired height
                        

                                        frame = cv2.resize(frame, (target_shape_w, target_shape_h)) # resize the image



                                        ## NOW, CONVERT IT FROM A 3D ARRAY TO 3 2D ARRAYS ##

                                        B, G, R = cv2.split(frame) # split to blue green red



                                        ## BIAS IT DOWN ##

                                        B = np.subtract(np.float32(B), 125) # bias it down

                                        G = np.subtract(np.float32(G), 125) # bias it down

                                        R = np.subtract(np.float32(R), 125) # bias it down


                                        ## FINALLY, CONVERT IT TO A REGULAR ARRAY ##

                                        B = B.tolist() # convert to string

                                        G = G.tolist() # convert to string

                                        R = R.tolist() # convert to string


                                        ## NOW WE RESHAPE IT TO FIT IN THE NEURON GROUPS ##

                                        big_array = [] # make a big array to hold all the information

                                        for z in B: # iterate through the blue channel

                                                for x in z: # iterate through the rows 

                                                        big_array.append(x) # append it to the array

                                        for z in G: # iterate through the green channel

                                                for x in z: # iterate through the rows 

                                                        big_array.append(x) # append it to the array

                                        for z in R: # iterate through the red channel

                                                for x in z: # iterate through the rows 

                                                        big_array.append(x) # append it to the array


                                else: # otherwise

                                        if random.randint(0,3) == 2: # if its good to go

                                                big_array = [] # make an array

                                                for z in range((int(self.prefs[0]) * int(self.prefs[1]) * int(self.prefs[3]) * int(self.prefs[4]))): # iterate through all the needed elements

                                                        big_array.append(125) # append a -70 to it



                                        else: # otherwise

                                                big_array = [] # make an array
                                                
                                                for z in range((int(self.prefs[0]) * int(self.prefs[1]) * int(self.prefs[3]) * int(self.prefs[4]))): # iterate through all the needed elements

                                                        big_array.append(-70) # append a -70 to it


                                for z in range((int(self.prefs[0]) * int(self.prefs[1]) * int(self.prefs[3]) * int(self.prefs[4])) - len(big_array)): # iterate through all the needed elements

                                        big_array.append(-70) # append a -70 to it


                                ## PACKAGE THE IMAGE INPUT DATA ##

                                video_input_layers = [] # make a list of video input layers to feed to the machine 

                                for z in range(int(self.prefs[3]) * int(self.prefs[4])): # iterate through the number of subarrays needed

                                        tmp = [] # make a tmp array

                                        for element in range(int(self.prefs[0]) * int(self.prefs[1])): # iterate through the index needed to populate the input arrays

                                                tmp.append(big_array[element + (z * int(self.prefs[0]) * int(self.prefs[1]))]) # populate the array

                                        video_input_layers.append(tmp) # append it to the output array


                                for z in range(int(self.prefs[3]) * int(self.prefs[4])): # iterate through the number of subarrays needed

                                        '''NOTE: THE FRONT INDEX IS WHERE WE ARE FEEDING IN THE VIDEO INPUT LAYER DATA'''

                                        NG_index = (z * int(self.prefs[5])) + int(self.prefs[5]) - 1 # calculate the front face index

                                        self.last_tick_interconnect_layers[NG_index][0] = video_input_layers[z] # update the input data layer


                                        

                        ## DETERMINE WHAT TO SEND TO THE PROCESSES ##

                        if mode != 'RUN': # if it's not a normal run session

                                commands = [] # make a list of commands to execute 

                                use_comamnd = False # set default command state


                                if mode == 'SAVE': # if we're to save the model


                                        use_command = True # use a command


                                        self.e_log.print_log('Simulation Is At {} ms, Attempting To Save The Simulation.'.format(str(self.tick_counter * int(self.prefs[11]) * float(self.prefs[7])))) # info stuff

                                        trymake(self.save_dir + '/AI_NEURAL_NET-{}-DIMS-{}-{}-{}/'.format(str(self.tick_counter), self.prefs[3], self.prefs[4], self.prefs[5])) # make the save folder


                                        with open(self.save_dir + '/AI_NEURAL_NET-{}-DIMS-{}-{}-{}/save_data.hiccup'.format(str(self.tick_counter), self.prefs[3], self.prefs[4], self.prefs[5]), 'w') as f: # open the file to write to it

                                                f.write(str(self.tick_counter)) # write the data



                                        for x in range(len(self.input_queues)): # iterate through the number of procs

                                                trymake(self.save_dir + '/AI_NEURAL_NET-{}/Process-{}'.format(str(self.tick_counter), str(x))) # make the save process folder

                                                commands.append('SAVE&{}/AI_NEURAL_NET-{}-DIMS-{}-{}-{}/Process-{}'.format(self.save_dir, str(self.tick_counter), self.prefs[3], self.prefs[4], self.prefs[5], str(x))) # append the command to the commands list


                                        self.e_log.print_log('Simulation Is At {} ms, Successfully Saved The Simulation.'.format(str(self.tick_counter * int(self.prefs[11]) * float(self.prefs[7])))) # info stuff





                                if mode == 'save_graph': # if we need to save the graph


                                        use_command = False # dont use a command


                                        self.e_log.print_log('Simulation Is At {} ms, Attempting To Save Info Stats.'.format(str(self.tick_counter * int(self.prefs[11]) * float(self.prefs[7])))) # info stuff


                                        ## CALCULATE VOLTAGE TOTALS ##

                                        v_tots = [] # set a voltage total

                                        for z in self.voltages: # iterate through the voltages


                                                v_tot = 0 # set inital value


                                                for y in z: # iterate through the subarrays

                                                        v_tot += y # add it


                                                v_tots.append(v_tot) # append it


                                 
                                        ## GENERATE THE GRAPH ##

                                        plt.cla() # clear the axes of the plot


                                        ntot = int(self.prefs[0]) * int(self.prefs[1]) * int(self.prefs[2]) * int(self.prefs[3]) * int(self.prefs[4]) * int(self.prefs[5]) # calculate total neurons
    
                                        plt.suptitle('AI (SNN-VARIANT-3) {} NEURONS [{}]'.format(str(ntot), str(datetime.datetime.now()))) # set graph title


                                        plt.xlabel('TIME (updates [Each Update Is {} ms])'.format(self.prefs[7])) # label the x axis

                                        plt.ylabel('VOLTAGE (mv)') # label the y axis


                                        plt.plot(v_tots) # make the graph

                                        plt.savefig(self.graph_path + '/AI_NEURAL_NET-{}-DIMS-{}-{}-{}.png'.format(str(self.tick_counter), self.prefs[3], self.prefs[4], self.prefs[5])) # save the graph





                                        ## DUMP THE RAW VOLTAGE TOTAL ARRAY DATA ##

                                        strout = '' # make raw outstring

                                        for z in v_tots: # iterate through the total voltages out

                                            strout += str(z) + ',' # make the csv

                                        with open(self.save_dir + '/vtots_data-{}N.csv'.format(str(ntot)), 'a') as f: # open the file with append

                                            f.write(strout) # write the outstring

                                        

                                        ## REINIT VOLTAGES ARRAY ##

                                        self.voltages = [] # clear the votlage array


                                        self.e_log.print_log('Simulation Is At {} ms, Saved Info Stats.'.format(str(self.tick_counter * int(self.prefs[11]) * float(self.prefs[7])))) # info stuff



                                elif mode == 'QUIT': # if it's to quit


                                        use_comamnd = True # use a command


                                        for x in range(len(self.input_queues)): # iterate through the number of procs

                                                commands.append('EXIT') # append the command to the commands list



                                ## SEND THE COMMAND IN THROUGH THE QUEUE ##

                                if use_command: # if we need to use a command


                                    for z in range(len(self.input_queues)): # iterate through the command to be sent

                                        self.input_queues[z].put(commands[z]) # put the command into the queue


                                
                        

                        ## INCRIMENT THE TICK COUNTER ##

                        self.tick_counter += 1 # incriment the tick counter



                        ## SEND THE DATA IN THROUGH THE QUEUE ##

                        for z in range(len(self.input_queues)): # iterate through the data to be sent

                                self.input_queues[z].put(self.last_tick_interconnect_layers[z]) # put the data into the queue



                        ## GET THE DATA BACK FROM THE OUTPUT QUEUE ##

                        self.last_tick_interconnect_layers = [] # clear the interconnect layer data

                        tmp = [] # make a tmp list

                        for z in range(len(self.output_queues)): # iterate through the data to be gotten

                                data = self.output_queues[z].get()

                                tmp.append(data[1]) # append the voltage totals

                                self.last_tick_interconnect_layers.append(data[0]) # get the data from the queue and put it into the array

                        self.voltages.append(tmp) # append the tmp list to the voltage list
                


                        ## CONNECT THE INTERCONNECT LAYERS ##

                        tmp_interconnect_layers = self.last_tick_interconnect_layers # make a temporary interconnect layer to connect with


                        ## NEXT, UPDATE THE VERTICAL INPUTS FOR EACH LAYER ##

                        for x in range(int(self.prefs[4]) - 1): # iterate through the layers

                                for w in range(int(self.prefs[3])): # iterate through the rows

                                        for y in range(int(self.prefs[5])): # iterate through the columns

                                                z = w * int(self.prefs[5]) # calculate z

                                                tmp_interconnect_layers[x * (int(self.prefs[3]) * int(self.prefs[4])) + z + y][3] = self.last_tick_interconnect_layers[(x + 1) * (int(self.prefs[3]) * int(self.prefs[4])) + z + y][3] # update the bottom side of the NGs

                                                tmp_interconnect_layers[x * (int(self.prefs[3]) * int(self.prefs[4])) + z + y][2] = self.last_tick_interconnect_layers[(x - 1) * (int(self.prefs[3]) * int(self.prefs[4])) + z + y][2] # update the top side of the NGs


                        ## FIRST, CONNECT THE OUTPUTS TO THE INPUTS ON EACH LAYER ##

                        for w in range(int(self.prefs[4])): # iterate through the layers

                                for z in range(int(self.prefs[3]) - 1): # iterate through the rows

                                        for y in range(int(self.prefs[5])): # iterate through the columns

                                                x = w * (int(self.prefs[4]) * int(self.prefs[5])) # calculate x

                                                tmp_interconnect_layers[x + z * int(self.prefs[5]) + y][5] = self.last_tick_interconnect_layers[x + (z + 1) * int(self.prefs[5]) + y][5] # update the right side of the NGs

                                                tmp_interconnect_layers[x + z * int(self.prefs[5]) + y][4] = self.last_tick_interconnect_layers[x + (z - 1) * int(self.prefs[5]) + y][4] # update the left side of the NGs


                        ## NEXT, UPDATE THE ADJACENT INPUTS FOR EACH LAYER ##

                        for q in range(int(self.prefs[4])): # iterate through the layers

                                for w in range(int(self.prefs[3])): # iterate through the rows

                                        for y in range(int(self.prefs[5]) - 1): # iterate through the columns

                                                x = q * (int(self.prefs[4]) * int(self.prefs[5])) # calculate x

                                                z = w * int(self.prefs[5]) # calculate z

                                                tmp_interconnect_layers[x + z + y][1] = self.last_tick_interconnect_layers[x + z + y + 1][1] # update the back side of the NGs

                                                tmp_interconnect_layers[x + z + y][0] = self.last_tick_interconnect_layers[x + z + y - 1][0] # update the front side of the NGs
        


                        ## UPDATE THE USER ##

                        if self.tick_counter % int(self.prefs[11]): # if it's to be updated

                                self.e_log.print_log('Simulation Is At {} ms.'.format(str(self.tick_counter * int(self.prefs[11]) * float(self.prefs[7])))) # print log the tick count



if __name__ == '__main__': # if it's the main module

        ASI_SNN = asi() # start the ai 

        ASI_SNN.run() # run the ai
